/**
 * Keyboard key names for event handling
 */
export enum KeyboardKey {
	ENTER = 'Enter',
	ESCAPE = 'Escape',
	ARROW_UP = 'ArrowUp',
	ARROW_DOWN = 'ArrowDown',
	ARROW_LEFT = 'ArrowLeft',
	ARROW_RIGHT = 'ArrowRight',
	TAB = 'Tab',
	B_LOWER = 'b',
	D_LOWER = 'd',
	D_UPPER = 'D',
	E_UPPER = 'E',
	K_LOWER = 'k',
	O_LOWER = 'o',
	O_UPPER = 'O',
	SPACE = ' '
}
